package com.dzeio.openhealth.data

import android.content.Context
import androidx.room.AutoMigration
import androidx.room.Database
import androidx.room.Room
import androidx.room.RoomDatabase
import com.dzeio.openhealth.data.food.Food
import com.dzeio.openhealth.data.food.FoodDao
import com.dzeio.openhealth.data.step.Step
import com.dzeio.openhealth.data.step.StepDao
import com.dzeio.openhealth.data.water.Water
import com.dzeio.openhealth.data.water.WaterDao
import com.dzeio.openhealth.data.weight.Weight
import com.dzeio.openhealth.data.weight.WeightDao
import java.io.File
import java.io.IOException

/**
 * ROOM SQLite database for the application
 *
 * It may be replaced if I want to fully encrypt the database
 */
@Database(
    entities = [
        Weight::class,
        Water::class,
        Step::class,
        Food::class
    ],
    version = 2,
    autoMigrations = [
        AutoMigration(1, 2)
    ],
    exportSchema = true
)
abstract class AppDatabase : RoomDatabase() {

    abstract fun weightDao(): WeightDao
    abstract fun waterDao(): WaterDao
    abstract fun stepDao(): StepDao

    abstract fun foodDao(): FoodDao

    companion object {
        /**
         * database name duh
         */
        private const val DATABASE_NAME = "open_health"

        private const val BACKUP_DATABASE = "-bkp"

        private const val SQLITE_SHMFILE_SUFFIX = "-wal"
        private const val SQLITE_WALFILE_SUFFIX = "-shm"

        // For Singleton instantiation
        @Volatile
        private var instance: AppDatabase? = null

        // get the Database instance
        fun getInstance(context: Context): AppDatabase {
            return instance ?: synchronized(this) {
                instance ?: buildDatabase(context).also { instance = it }
            }
        }

        /**
         * build teh database
         */
        private fun buildDatabase(context: Context): AppDatabase {
            return Room.databaseBuilder(context, AppDatabase::class.java, DATABASE_NAME)
//                .addMigrations(MIGRATION_2_3)
                .build()
        }
    }

    fun backup(context: Context): Int {
        var result = -99
        if (instance == null) return result

        val dbFile = context.getDatabasePath(DATABASE_NAME)
        val dbWalFile = File(dbFile.path + SQLITE_WALFILE_SUFFIX)
        val dbShmFile = File(dbFile.path + SQLITE_SHMFILE_SUFFIX)
        val bkpFile = File(dbFile.path + BACKUP_DATABASE)
        val bkpWalFile = File(bkpFile.path + SQLITE_WALFILE_SUFFIX)
        val bkpShmFile = File(bkpFile.path + SQLITE_SHMFILE_SUFFIX)
        if (bkpFile.exists()) bkpFile.delete()
        if (bkpWalFile.exists()) bkpWalFile.delete()
        if (bkpShmFile.exists()) bkpShmFile.delete()
        checkpoint()
        try {
            dbFile.copyTo(bkpFile, true)
            if (dbWalFile.exists()) dbWalFile.copyTo(bkpWalFile, true)
            if (dbShmFile.exists()) dbShmFile.copyTo(bkpShmFile, true)
            result = 0
        } catch (e: IOException) {
            e.printStackTrace()
        }
        return result
    }

    fun restore(context: Context) {
        if (!File(context.getDatabasePath(DATABASE_NAME).path + BACKUP_DATABASE).exists()) {
            return
        }
        if (instance == null) return
        val dbpath = instance!!.openHelper.readableDatabase.path
        val dbFile = File(dbpath)
        val dbWalFile = File(dbFile.path + SQLITE_WALFILE_SUFFIX)
        val dbShmFile = File(dbFile.path + SQLITE_SHMFILE_SUFFIX)
        val bkpFile = File(dbFile.path + BACKUP_DATABASE)
        val bkpWalFile = File(bkpFile.path + SQLITE_WALFILE_SUFFIX)
        val bkpShmFile = File(bkpFile.path + SQLITE_SHMFILE_SUFFIX)
        try {
            bkpFile.copyTo(dbFile, true)
            if (bkpWalFile.exists()) bkpWalFile.copyTo(dbWalFile, true)
            if (bkpShmFile.exists()) bkpShmFile.copyTo(dbShmFile, true)
            checkpoint()
        } catch (e: IOException) {
            e.printStackTrace()
        }
    }

    private fun checkpoint() {
        val db = this.openHelper.writableDatabase
        db.query("PRAGMA wal_checkpoint(FULL);", emptyArray())
        db.query("PRAGMA wal_checkpoint(TRUNCATE);", emptyArray())
    }
}
